@@ -5,12 +5,14 @@ import base64
 import requests
 import jwcrypto.jwk as jwk
 
+from django.apps import apps
 from django.contrib import auth as django_auth
 from django.contrib.auth.models import AnonymousUser
 from django.core.exceptions import PermissionDenied
 from django.db.models import Q
 from django.core.exceptions import MultipleObjectsReturned
 from django.shortcuts import redirect
+from django.urls import reverse
 from rest_framework.authentication import BaseAuthentication
 from rest_framework import exceptions
 
@@ -18,12 +20,48 @@ from dbmi_client.settings import dbmi_settings
 from dbmi_client import authz
 
 # Get the app logger
-logger = dbmi_settings.get_logger()
+import logging
+logger = logging.getLogger(dbmi_settings.LOGGER_NAME)
 
 # Set a key to cache JWKs under in the DBMI.AUTH0 settings
 CACHED_JWKS_KEY = '__DBMI_CLIENT_CACHED_JWKS__'
 
 
+def login(request):
+    """
+    Builds and returns a redirect response to allow the user to login and send them to the
+    supplied URL when successfully logged in. If not defined in LOGIN_REDIRECT_URL, the
+    current URL in the request will be used, so the user will return to the original location.
+    :param request: The original request object
+    :return: The redirect response
+    """
+    return login_redirect(request)
+
+
+def login_redirect(request, next_url=None):
+    """
+    Builds and returns a redirect response to allow the user to login and send them to the
+    supplied URL when successfully logged in. If next_url is not passed, and
+    it is not defined in LOGIN_REDIRECT_URL, the current URL in the request will be used,
+    so the user will return to the original location.
+    :param request: The original request object
+    :param next_url: The URL users will be sent after login
+    :return: Response
+    """
+
+    # Ensure the request is cleared of user state
+    django_auth.logout(request)
+
+    # Get the url
+    login_url = login_redirect_url(request, next_url)
+
+    # Just process the logout and redirect them
+    response = redirect(login_url)
+
+    # Do needed logout functions and return the modified response
+    return response
+
+
 def login_redirect_url(request, next_url=None):
     """
     Builds and returns a URL that sends the user to the login, and returns them to the
@@ -33,17 +71,33 @@ def login_redirect_url(request, next_url=None):
     :param next_url: The URL users will be sent after login
     :return: Response
     """
+    # Check for local login enabled
+    if apps.is_installed('dbmi_client.login'):
 
-    # Build the URL
-    login_url = furl(dbmi_settings.AUTHN_URL)
-    login_url.path.segments.extend(['login', 'auth'])
-
-    # Check for the next URL
-    if next_url:
-        login_url.query.params.add('next', next_url)
+        # Use local login URL
+        login_url = furl(request.build_absolute_uri(reverse('dbmi_login:login')))
 
     else:
-        login_url.query.params.add('next', request.build_absolute_uri())
+
+        # Build the URL using DBMI-AuthN
+        login_url = furl(dbmi_settings.AUTHN_URL)
+        login_url.path.segments.extend(['login', 'auth'])
+
+    # If no next URL, determine where to dump them after logout
+    if not next_url:
+        if dbmi_settings.LOGIN_REDIRECT_URL:
+            next_url = request.build_absolute_uri(dbmi_settings.LOGIN_REDIRECT_URL)
+
+        else:
+            next_url = request.build_absolute_uri()
+
+    # Add next url
+    logger.debug('Login next URL: {}'.format(next_url))
+    login_url.query.params.add(dbmi_settings.LOGIN_REDIRECT_KEY, next_url)
+
+    # Add the default client ID, if specified
+    if hasattr(dbmi_settings, 'AUTH0_CLIENT_ID') and getattr(dbmi_settings, 'AUTH0_CLIENT_ID'):
+        login_url.query.params.add('client_id', dbmi_settings.AUTH0_CLIENT_ID)
 
     # Check for branding
     if dbmi_settings.AUTHN_TITLE or dbmi_settings.AUTHN_ICON_URL:
@@ -60,28 +114,92 @@ def login_redirect_url(request, next_url=None):
         branding_param = base64.urlsafe_b64encode(json.dumps(branding).encode('utf-8')).decode('utf-8')
         login_url.query.params.add('branding', branding_param)
 
+    logger.debug('Login URL: {}'.format(login_url.url))
     return login_url.url
 
 
+def logout(request):
+    """
+    This will prepare the redirect to log the user out (either internally or at DBMI-AuthN)
+    :param request: The Django request object
+    :return: The response object that takes the user to the logout endpoint
+    """
+    # Call logout redirect
+    return logout_redirect(request)
+
+
 def logout_redirect(request, next_url=None):
     """
-    This will log a user out and redirect them to log in again via the AuthN server.
-    :param request: The Django request object
-    :param next_url: A URL that the user should be sent to should they log back in
-    :return: The response object that takes the user to the login page. 'next' parameter set to bring them back to their intended page.
+    This is just an alias for `logout` to support older clients
     """
     # Ensure the request is cleared of user state
     django_auth.logout(request)
 
-    # Get a login response
-    response = redirect(login_redirect_url(request, next_url))
+    # Build the logout URL
+    logout_url = logout_redirect_url(request, next_url)
 
-    # Set the URL and purge cookies
-    response.delete_cookie(dbmi_settings.JWT_COOKIE_NAME, domain=dbmi_settings.JWT_COOKIE_DOMAIN)
+    # Just process the logout and redirect them
+    response = redirect(logout_url)
 
+    # Do needed logout functions and return the modified response
     return response
 
 
+def logout_redirect_url(request, next_url=None):
+    """
+    This will prepare the redirect URL to log the user out (either internally or at DBMI-AuthN)
+    :param request: The Django request object
+    :param next_url: A URL that the user should be sent to should they log back in
+    :return: The response object that takes the user to the logout endpoint
+    """
+    # Check for local login enabled
+    if apps.is_installed('dbmi_client.login'):
+
+        # Build the URL to DBMI-Client's logout page
+        logout_url = furl(request.build_absolute_uri(reverse('dbmi_login:logout')))
+
+    else:
+
+        # Build the URL to DBMI-AuthN's logout endpoint
+        logout_url = furl(dbmi_settings.AUTHN_URL)
+        logout_url.path.segments.extend(['login', 'logout'])
+
+    # If no next URL, determine where to dump them after logout
+    if not next_url:
+        if dbmi_settings.LOGOUT_REDIRECT_URL:
+            next_url = request.build_absolute_uri(dbmi_settings.LOGOUT_REDIRECT_URL)
+
+        else:
+            next_url = request.build_absolute_uri()
+
+    logger.debug('Logout next URL: {}'.format(next_url))
+
+    # Add next url
+    logout_url.query.params.add(dbmi_settings.LOGOUT_REDIRECT_KEY, next_url)
+
+    # Add the default client ID, if specified
+    if hasattr(dbmi_settings, 'AUTH0_CLIENT_ID') and getattr(dbmi_settings, 'AUTH0_CLIENT_ID'):
+        logout_url.query.params.add('client_id', dbmi_settings.AUTH0_CLIENT_ID)
+
+    # Check for branding
+    if dbmi_settings.AUTHN_TITLE or dbmi_settings.AUTHN_ICON_URL:
+
+        # Add the included parameters
+        branding = {}
+        if dbmi_settings.AUTHN_TITLE:
+            branding['title'] = dbmi_settings.AUTHN_TITLE
+
+        if dbmi_settings.AUTHN_TITLE:
+            branding['icon_url'] = dbmi_settings.AUTHN_ICON_URL
+
+        # Encode it and pass it along
+        branding_param = base64.urlsafe_b64encode(json.dumps(branding).encode('utf-8')).decode('utf-8')
+        logout_url.query.params.add('branding', branding_param)
+
+    logger.debug('Logout URL: {}'.format(logout_url.url))
+    return logout_url.url
+
+
 def dbmi_http_headers(request, content_type='application/json', **kwargs):
     """
     Returns headers to be used for API calls to DBMI services in order to authenticate the caller
@@ -165,19 +283,41 @@ def validate_request(request):
 
     # Extract JWT token into a string.
     jwt_string = get_jwt(request)
+    algorithm = None
 
-    # Check that we actually have a token.
-    if jwt_string is not None:
-        return validate_rs256_jwt(jwt_string)
-    else:
+    # Ensure JWT exists
+    if not jwt_string:
         return None
 
+    # Determine which Auth0 Client ID (aud) this JWT pertains to.
+    try:
+        unverified_header = jwt.get_unverified_header(str(jwt_string))
+        algorithm = unverified_header.get('alg', 'rs256').lower()
 
-def get_public_keys_from_auth0(refresh=False):
+        # Check algorithm
+        if algorithm == 'hs256':
+            return validate_hs256_jwt(jwt_string)
+
+        elif algorithm == 'rs256':
+            return validate_rs256_jwt(jwt_string)
+
+        else:
+            logger.warning(f'Unsupported JWT algorithm: {algorithm}')
+            return None
+
+    except Exception as e:
+        logger.exception('Validate error: {}'.format(e), exc_info=True,
+                         extra={'algorithm': algorithm})
+
+    return None
+
+
+def get_public_keys_from_auth0(tenant, refresh=False):
     '''
     Retrieves the public key from Auth0 to verify JWTs. Will
     cache the JSON response from Auth0 in Django settings
     until instructed to refresh the JWKS.
+    :param tenant: The Auth0 tenant to fetch JWKs for
     :param refresh: Purges cached JWK and fetches from remote
     :return: dict
     '''
@@ -185,21 +325,21 @@ def get_public_keys_from_auth0(refresh=False):
     # If refresh, delete cached key
     if refresh:
         logger.debug('Refresh requested, deleting cached JWKs')
-        delattr(dbmi_settings, CACHED_JWKS_KEY)
+        delattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}')
 
     try:
         # Look in settings
-        if hasattr(dbmi_settings, CACHED_JWKS_KEY):
+        if hasattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}'):
 
             # Parse the cached dict and return it
-            return json.loads(getattr(dbmi_settings, CACHED_JWKS_KEY))
+            return json.loads(getattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}'))
 
         else:
 
             logger.debug('Fetching remote JWKS')
 
             # Build the JWKs URL
-            url = furl().set(scheme='https', host='{}.auth0.com'.format(dbmi_settings.AUTH0_TENANT))
+            url = furl().set(scheme='https', host='{}.auth0.com'.format(tenant))
             url.path.segments.extend(['.well-known', 'jwks.json'])
 
             # Make the request
@@ -210,23 +350,23 @@ def get_public_keys_from_auth0(refresh=False):
             jwks = response.json()
 
             # Cache it
-            setattr(dbmi_settings, CACHED_JWKS_KEY, json.dumps(jwks))
+            setattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}', json.dumps(jwks))
 
             return jwks
 
     except KeyError as e:
-        logger.exception(e)
+        logger.exception(f'Error getting public keys: {e}', exc_info=True, extra={'tenant': tenant})
 
     except json.JSONDecodeError as e:
-        logger.exception(e)
+        logger.exception(f'Error getting public keys: {e}', exc_info=True, extra={'tenant': tenant})
 
     except requests.HTTPError as e:
-        logger.exception(e)
+        logger.exception(f'Error getting public keys: {e}', exc_info=True, extra={'tenant': tenant})
 
     return None
 
 
-def retrieve_public_key(jwt_string):
+def retrieve_public_key(tenant, jwt_string):
     '''
     Gets the public key used to sign the JWT from the public JWK
     hosted on Auth0. Auth0 typically only returns one public key
@@ -239,16 +379,17 @@ def retrieve_public_key(jwt_string):
     fresh JWKS is then searched again for the needed key.
 
     Returns the key ID if found, otherwise returns None
+    :param tenant: The Auth0 tenant to fetch JWKs for
     :param jwt_string: The JWT token as a string
     :return: str
     '''
 
     try:
         # Get the JWK
-        jwks = get_public_keys_from_auth0(refresh=False)
+        jwks = get_public_keys_from_auth0(tenant, refresh=False)
         if not jwks:
             logger.debug('Could not fetch JWKs from Auth0, just fail out now')
-            raise PermissionDenied
+            return None
 
         # Decode the JWTs header component
         unverified_header = jwt.get_unverified_header(str(jwt_string))
@@ -261,21 +402,19 @@ def retrieve_public_key(jwt_string):
             logger.debug('Cached JWK keys: {}'.format([jwk['kid'] for jwk in jwks['keys']]))
 
             # No match found, refresh the jwks
-            jwks = get_public_keys_from_auth0(refresh=True)
+            jwks = get_public_keys_from_auth0(tenant, refresh=True)
             logger.debug('Refreshed JWK keys: {}'.format([jwk['kid'] for jwk in jwks['keys']]))
 
             # Try it again
             rsa_key = get_rsa_from_jwks(jwks, unverified_header['kid'])
             if not rsa_key:
-                logger.warning('Invalid JWT attempt', extra={
-                    'unverified_kid': unverified_header['kid'], 'jwt': jwt_string,
-                })
-                raise PermissionDenied
+                logger.warning('Invalid JWT attempt', extra={'unverified_kid': unverified_header['kid']})
+                return None
 
         return rsa_key
 
     except KeyError as e:
-        logger.debug('Could not compare keys, probably old HS256 session: {}'.format(e))
+        logger.exception('Error retrieving public keys: {}'.format(e), exc_info=True, extra={'tenant': tenant})
 
     return None
 
@@ -313,24 +452,68 @@ def validate_rs256_jwt(jwt_string):
     :param jwt_string: JWT as a string
     :return: dict
     '''
+    # We need the public key and the client ID to match against
+    jwk_pub_key = None
+    jwt_client_id = None
 
-    rsa_pub_key = retrieve_public_key(jwt_string)
-    if rsa_pub_key:
+    # Determine which Auth0 Client ID (aud) this JWT pertains to.
+    try:
+        jwt_client_id = str(jwt.decode(jwt_string, verify=False)['aud'])
+
+        # Check if multiple clients are specified at the client level
+        if hasattr(dbmi_settings, 'AUTH0_CLIENTS') and getattr(dbmi_settings, 'AUTH0_CLIENTS'):
+
+            # Search client IDs
+            tenant = dbmi_settings.AUTH0_CLIENTS.get(jwt_client_id, {}).get('tenant')
+            if tenant:
+                logger.debug(f'JWT Client ID matched to tenant: {tenant}')
+
+                # Get the public key
+                jwk_pub_key = retrieve_public_key(tenant, jwt_string)
+
+            # Log if not found
+            else:
+                logger.info(f'JWT client {jwt_client_id} could not be matched to any clients')
+
+        # Check for multiple clients at the tenant level
+        if not jwk_pub_key and hasattr(dbmi_settings, 'AUTH0_TENANTS') and getattr(dbmi_settings, 'AUTH0_TENANTS'):
+
+            # Try each one
+            for tenant in dbmi_settings.AUTH0_TENANTS:
+                try:
+                    # Get the public key
+                    jwk_pub_key = retrieve_public_key(tenant, jwt_string)
+
+                    # A public key was returned, we've got a JWT from this tenant
+                    if jwk_pub_key:
+                        break
+
+                except PermissionDenied:
+                    pass
+            else:
+                logger.info(f'JWT client {jwt_client_id} could not be matched to '
+                            f'any tenants: {dbmi_settings.AUTH0_TENANTS}')
+
+        # If not already matched, try default client
+        if not jwk_pub_key and jwt_client_id == dbmi_settings.AUTH0_CLIENT_ID:
+
+            # Get the public key
+            jwk_pub_key = retrieve_public_key(dbmi_settings.AUTH0_TENANT, jwt_string)
+
+        if not jwk_pub_key:
+            logger.warning(f'JWT Client ID could not be matched: {jwt_client_id}')
+            return None
+
+    except Exception as e:
+        logger.exception(f'Failed to get the aud from jwt payload: {e}', exc_info=True,
+                         extra={'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key})
+        return None
+
+    # Attempt to validate with each one
+    if jwk_pub_key:
 
         # Get the JWK
-        jwk_key = jwk.JWK(**rsa_pub_key)
-
-        # Determine which Auth0 Client ID (aud) this JWT pertains to.
-        try:
-            auth0_client_id = str(jwt.decode(jwt_string, verify=False)['aud'])
-        except Exception as e:
-            logger.error('Failed to get the aud from jwt payload: {}'.format(e))
-            return None
-
-        # Check that the Client ID is in the allowed list of Auth0 Client IDs for this application
-        if not auth0_client_id == dbmi_settings.AUTH0_CLIENT_ID:
-            logger.error('Auth0 Client ID not allowed')
-            return None
+        jwk_key = jwk.JWK(**jwk_pub_key)
 
         # Attempt to validate the JWT (Checks both expiry and signature)
         try:
@@ -338,18 +521,69 @@ def validate_rs256_jwt(jwt_string):
                                  jwk_key.export_to_pem(private_key=False),
                                  algorithms=['RS256'],
                                  leeway=120,
-                                 audience=auth0_client_id)
+                                 audience=jwt_client_id)
 
             return payload
 
         except jwt.ExpiredSignatureError as e:
-            logger.warning("JWT Expired: {}".format(e))
+            logger.debug("JWT Expired: {}".format(e), extra={
+                'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key
+            })
 
         except jwt.InvalidTokenError as e:
-            logger.warning("Invalid JWT Token: {}".format(e))
+            logger.info("Invalid JWT Token: {}".format(e), extra={
+                'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key
+            })
 
         except Exception as e:
-            logger.error("Exception: {}".format(e))
+            logger.exception(f'Error validating JWT: {e}', exc_info=True, extra={
+                'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key
+            })
+
+    return None
+
+
+def validate_hs256_jwt(jwt_string):
+    '''
+    Verifies the given HS256 JWT. Returns the payload
+    if verified, otherwise returns None.
+    :param jwt_string: JWT as a string
+    :return: dict
+    '''
+
+    # Check settings for proper setup
+    if not dbmi_settings.AUTH0_SECRET or not dbmi_settings.AUTH0_CLIENT_ID:
+        logger.error('Cannot verify HS256 tokens without client ID and client secret')
+        raise PermissionDenied
+
+    # Determine which Auth0 Client ID (aud) this JWT pertains to.
+    jwt_client_id = None
+    try:
+        jwt_client_id = str(jwt.decode(jwt_string, verify=False)['aud'])
+
+        # Perform the validation
+        payload = jwt.decode(jwt_string,
+                             base64.b64decode(dbmi_settings.AUTH0_SECRET, '-_'),
+                             algorithms=['HS256'],
+                             leeway=120,
+                             audience=jwt_client_id)
+
+        return payload
+
+    except jwt.ExpiredSignatureError as e:
+        logger.debug("JWT Expired: {}".format(e), extra={
+            'jwt_client_id': jwt_client_id,
+        })
+
+    except jwt.InvalidTokenError as e:
+        logger.info("Invalid JWT Token: {}".format(e), extra={
+            'jwt_client_id': jwt_client_id,
+        })
+
+    except Exception as e:
+        logger.exception(f'Error validating JWT: {e}', exc_info=True, extra={
+            'jwt_client_id': jwt_client_id,
+        })
 
     return None
 
@@ -362,6 +596,8 @@ def validate_rs256_jwt(jwt_string):
 
 class DBMIAuthenticationBackend(object):
 
+    request = None
+
     def authenticate(self, request, token=None):
         """
         All versions of this backend follow the same flow:
@@ -401,12 +637,34 @@ class DBMIAuthenticationBackend(object):
             # Sync the user to ensure we've got updated properties
             self._sync_user(request, user)
 
+        # Retain request for permissions checks
+        self.request = request
+
         return user
 
     def get_user(self, user_id):
         # Should be implemented by subclass depending on data source for user
         raise SystemError('This method should not be called')
 
+    def has_perm(self, user, perm, obj):
+        """
+        Returns whether the given user has the permission or not
+        """
+        logger.debug(f'has_perm: {user} - {perm} - {obj}')
+
+        # Return permission
+        return authz.has_permission(request=self.request, email=user.email, item=obj, permission=perm, check_parents=True)
+
+    def has_module_perms(self, user, app_label):
+        """
+        Returns whether the given user has permissions on the module or not
+        """
+        logger.debug(f'has_module_perms: {user} - {app_label}')
+
+        # Return permission
+        permissions = authz.get_permissions(request=self.request, email=user.email, item=app_label)
+        return permissions and len(permissions)
+
     def _get_user_object(self, request):
         """
         Accepts details from the JWT user and returns an object representing
@@ -633,6 +891,36 @@ class DBMIModelAuthenticationBackend(DBMIAuthenticationBackend):
                              extra={'user': user.id, 'request': request})
 
 
+class DBMIUsersModelAuthenticationBackend(DBMIModelAuthenticationBackend):
+
+    """
+    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
+    Use this authentication backend for sites that are accessible to all users as well as administrator users.
+    User model is keyed by the username and email contained in the JWT. `is_staff` and `is_superuser` flags
+    are automatically set and synced on appropriate users. Every active user with admin authorization on this site
+    will have complete access to everything.
+    """
+
+    def _sync_user(self, request, user):
+        """
+        Called after a user is fetched/created and syncs any additional properties
+        from the JWT's payload to the user object. Set staff and superuser flags
+        if authorizations are valid.
+        """
+        # Do normal sync first
+        super(DBMIModelAuthenticationBackend, self)._sync_user(request, user)
+
+        # Check if admin
+        is_admin = authz.is_admin(request, user.email)
+        if is_admin:
+            logger.debug(f'User: {user.email} has been granted admin/superuser privileges')
+
+        # Ensure the model is updated
+        user.is_staff = is_admin
+        user.is_superuser = is_admin
+        user.save()
+
+
 class DBMIAdminModelAuthenticationBackend(DBMIModelAuthenticationBackend):
 
     """
@@ -697,7 +985,7 @@ class DBMISuperuserModelAuthenticationBackend(DBMIAdminModelAuthenticationBacken
         if authorizations are valid.
         """
         # Do normal sync first
-        super(DBMIAdminModelAuthenticationBackend, self)._sync_user(request, user)
+        super(DBMISuperuserModelAuthenticationBackend, self)._sync_user(request, user)
 
         # Check if admin
         if is_admin is None:
@@ -733,11 +1021,13 @@ class DBMIJWTUser(AnonymousUser):
     is_active = True
     is_staff = False
     is_superuser = False
+    jwt = None
 
     def __init__(self, request):
 
         # Get the payload
         payload = get_jwt_payload(request, verify=False)
+        self.jwt = get_jwt(request)
 
         # Set properties
         self.username = payload.get('sub')
@@ -786,7 +1076,8 @@ class DBMIJWTUser(AnonymousUser):
         return permissions
 
     def get_all_permissions(self, obj=None):
-        return _user_get_all_permissions(self, obj)
+        # Get permissions from AuthZ
+        return authz.get_permissions(self.jwt, self.email, obj)
 
     def has_perm(self, perm, obj=None):
         """
@@ -822,27 +1113,18 @@ class DBMIJWTUser(AnonymousUser):
         return _user_has_module_perms(self, app_label)
 
 
-# A few helper functions for common logic between User and AnonymousUser.
-def _user_get_all_permissions(user, obj):
-    permissions = set()
-
-    # TODO: Query DBMI AuthZ and return all permission items 'item.permission'
-
-    return permissions
-
-
 def _user_has_perm(user, perm, obj):
     """
     A backend can raise `PermissionDenied` to short-circuit permission checking.
     """
-    # Get all perms and compare
-    try:
-        for permission in _user_get_all_permissions(user, obj):
-            if permission == '{}.{}'.format(obj, perm):
+    for backend in django_auth.get_backends():
+        if not hasattr(backend, 'has_perm'):
+            continue
+        try:
+            if backend.has_perm(user, perm, obj):
                 return True
-    except PermissionDenied:
-        return False
-
+        except PermissionDenied:
+            return False
     return False
 
 
@@ -850,14 +1132,14 @@ def _user_has_module_perms(user, app_label):
     """
     A backend can raise `PermissionDenied` to short-circuit permission checking.
     """
-    # Get all perms and compare
-    try:
-        for perm in _user_get_all_permissions(user, app_label):
-            if '{}.{}'.format(app_label, perm) == perm:
+    for backend in django_auth.get_backends():
+        if not hasattr(backend, 'has_module_perms'):
+            continue
+        try:
+            if backend.has_module_perms(user, app_label):
                 return True
-    except PermissionDenied:
-        return False
-
+        except PermissionDenied:
+            return False
     return False
 
 ###################################################################
